from logging import basicConfig, config, WARNING
from os import path
from sys import exit
from kickstart_iso_tool.constants import (
    ISOINFO_BIN,
    XORRISO,
)


if not path.exists(ISOINFO_BIN):
    print(f"{ISOINFO_BIN} doesn't exists, try to install -> genisoimage")
    exit(1)
if not path.exists(XORRISO):
    print(f"{XORRISO} doesn't exists, try to install -> xorriso")
    exit(1)

try:
    config.fileConfig("/etc/kiso/logging.ini")
except KeyError:
    basicConfig(level=WARNING)
except FileNotFoundError:
    basicConfig(level=WARNING)


import logging
import os
import subprocess
import re
import pycdlib
import collections
import jinja2
from .constants import (
    ISOINFO_BIN,
    XORRISO,
)
from .base import (
    execute_command,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

def create_grub_cfg(iso_name: str, template:str, name: str ) -> str:
    template = jinja2.Template(template)
    return template.render(name=name, iso_name=iso_name)

def iso_volume_id(iso_location) -> str:
    print(f"executing: {ISOINFO_BIN} -d -i {iso_location}")
    command = [ISOINFO_BIN, "-d", "-i", iso_location]
    exec = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    output, err = exec.communicate()
    iso_descriptor = output.decode("UTF-8")
    iso_descriptor = iso_descriptor.splitlines()
    for i in range(len(iso_descriptor)):
        line = iso_descriptor[i]
        if re.match("Volume id", line):
            volume_id = line.split(" ")[-1]
            logger.debug(f"Volume id: {volume_id}")
            return volume_id

def create_kickstarted_iso(
    src: str,
    dst: str,
    iso_name: str,
    kickstart_cfg: str,
    grub_cfg: str,
    default_boot_cfg: str,
    default_boot_file: str,
    optional_kickstart_cfg: str = None,
) -> None:
    logger.info(f"Creating kickstarted iso image at {dst}")
    command = [
        XORRISO, "-indev", src,
        "-outdev", dst,
        "-boot_image", "any", "replay",
        "-joliet", "on",
        "-system_id", "LINUX",
        "-volid", iso_name,
        "-map", kickstart_cfg, "ks.cfg",
        "-map", grub_cfg, "EFI/BOOT/grub.cfg",
        "-map", default_boot_cfg, default_boot_file,
    ]
    if optional_kickstart_cfg:
        logger.debug(f"Adding optional kickstart configuration: {optional_kickstart_cfg}")
        base_name = os.path.basename(optional_kickstart_cfg)
        command.append("-map")
        command.append(optional_kickstart_cfg)
        command.append(base_name)
    logger.debug("Executing command: %s", ' '.join(command))
    execute_command(command)


HUMAN_UNITS = ["B", "KiB", "MiB", "GiB", "TiB", "PiB"]
SPEED_UNITS = ["bps", "Kbps", "Mbps", "Gbps", "Tbps"]
CHUNK_SIZE = 1024
ISOINFO_BIN = "/usr/bin/isoinfo"
XORRISO = "/usr/bin/xorriso"
grub_cfg_template = """
set default="0"
function load_video {
  insmod all_video
}
load_video
set gfxpayload=keep
insmod gzio
insmod part_gpt
insmod ext2
insmod chain
set timeout=5
search --no-floppy --set=root -l '{{ iso_name }}'
menuentry 'Kickstarted install of {{ name }}' --class fedora --class gnu-linux --class gnu --class os {
	linux /images/pxeboot/vmlinuz inst.stage2=hd:LABEL={{ iso_name }} inst.ks=hd:LABEL={{ iso_name }} net.ifnames=0 biosdevname=0 mitigations=off inst.text
	initrd /images/pxeboot/initrd.img
}
"""
isolinux_cfg_template = """
default vesamenu.c32
timeout 100
display boot.msg
menu clear
menu background splash.png
menu title Rocky Linux 9.5
menu vshift 8
menu rows 18
menu margin 8
menu helpmsgrow 15
menu tabmsgrow 13
menu color border * #00000000 #00000000 none
menu color sel 0 #ffffffff #00000000 none
menu color title 0 #ff7ba3d0 #00000000 none
menu color tabmsg 0 #ff3a6496 #00000000 none
menu color unsel 0 #84b8ffff #00000000 none
menu color hotsel 0 #84b8ffff #00000000 none
menu color hotkey 0 #ffffffff #00000000 none
menu color help 0 #ffffffff #00000000 none
menu color scrollbar 0 #ffffffff #ff355594 none
menu color timeout 0 #ffffffff #00000000 none
menu color timeout_msg 0 #ffffffff #00000000 none
menu color cmdmark 0 #84b8ffff #00000000 none
menu color cmdline 0 #ffffffff #00000000 none
menu tabmsg Press Tab for full configuration options on menu items.
menu separator # insert an empty line
menu separator # insert an empty line
label linux
  menu label ^Kickstarted install of Rocky Linux {{ name }}
  menu default
  kernel vmlinuz
  append initrd=initrd.img inst.stage2=hd:LABEL={{ iso_name }} inst.ks=hd:LABEL={{ iso_name }} net.ifnames=0 biosdevname=0 mitigations=off inst.text
"""


import requests
from urllib.parse import urlparse
from logging import getLogger
import sys
import time
import os
import shutil
import subprocess
from kickstart_iso_tool.constants import (
    SPEED_UNITS,
    CHUNK_SIZE,
    HUMAN_UNITS
)

logger = getLogger(__name__)


def convert_size_to_human(size: float, speed=None):
    if speed:
        for unit in SPEED_UNITS:
            if unit == SPEED_UNITS[-1]:
                logger.warning(f"Breaking from loop as reached {SPEED_UNITS[-1]}")
                break
            if float(size) < 1024.0:
                break
            size = float(size) / 1024.0
        size *= 8
    else:
        for unit in HUMAN_UNITS:
            if unit == HUMAN_UNITS[-1]:
                logger.warning(f"Breaking from loop as reached {HUMAN_UNITS[-1]}")
                break
            if float(size) < 1024.0:
                break
            size = float(size) / 1024.0
    return f"{size:.2f} {unit}"


def progress_bar(terminal_size: int, current_length: int, total_length: int, speed: float, file_name: str) -> None:
    done_percent = round(current_length * 100 / total_length, 2)
    bar_prefix = f"Downloading {file_name}:"
    bar_suffix = f"{done_percent:0>6.2f}% | Speed: {speed}"
    bar_length = (len(bar_prefix) + len(bar_suffix)) + 1
    bar_size = terminal_size - bar_length - 3
    done = int(bar_size * current_length / total_length)
    done_length = "=" * (done)
    in_progress_length = "." * (bar_size - done - 1)
    if bar_size <= done:
        current_progress = ""
    else:
        current_progress = ">"
    if bar_length > terminal_size:
        sys.stdout.write(f"\rDL: {done_percent:0>6.2f}%")
    else:
        sys.stdout.write(f"\r{bar_prefix} [{done_length}{current_progress}{in_progress_length}] {bar_suffix}")


def download_file(url: str, save_path: str):
    logger.debug(f"URL: {url}")
    logger.debug(f"Download save path: {save_path}")
    file_name = urlparse(url)
    file_name = os.path.basename(file_name.path)
    start_time = time.perf_counter()
    download = requests.get(url, stream=True)
    total_length = download.headers.get("content-length")
    if os.path.exists(save_path):
        logger.warning(f"File: {save_path} already exists")
        old_file_size = os.stat(save_path)
        logger.debug(f"Old file size: {convert_size_to_human(old_file_size.st_size)}")
        if int(total_length) > int(old_file_size.st_size):
            logger.warning("Remote file size different, re-downloading")
        else:
            logger.warning("Old file size matches remote, re-using!")
            return save_path
    with open(save_path, "wb") as file:
        check_free_space(int(total_length) * 2, save_path)
        totoal_length_human = convert_size_to_human(total_length)
        logger.debug(f"Download size: {totoal_length_human}")
        current_length = 0
        if total_length is None:
            file.write(download.content)
        else:
            total_length = int(total_length)
            dl_speed = 0
            update_time = 1
            for data in download.iter_content(chunk_size=CHUNK_SIZE):
                TERMINAL_SIZE = os.get_terminal_size().columns
                file.write(data)
                current_length += len(data)
                current_time = time.perf_counter()
                time_diff = (current_time - start_time)
                if int(time_diff) + 1 == update_time:
                    dl_speed = convert_size_to_human(current_length / time_diff, speed=True)
                    update_time += 1
                    progress_bar(TERMINAL_SIZE, current_length, total_length, dl_speed, file_name=file_name)
            progress_bar(TERMINAL_SIZE, current_length, total_length, dl_speed, file_name=file_name)
    return save_path


def check_free_space(size: int, save_path: str) -> None:
    current_usage = shutil.disk_usage(save_path)
    free = current_usage.free
    size = round(size * 1.1)
    logger.debug(f"Available disk space: {convert_size_to_human(free)} for path: {save_path}")
    logger.debug(f"Estimated needs: {convert_size_to_human(size)}")
    if size >= free:
        logger.critical(f"Low disk space for {save_path}")
        logger.critical(f"Available space: {convert_size_to_human(free)}")
        logger.critical(f"Expected usage: {convert_size_to_human(size)}")
        sys.exit(1)


def execute_command(shell_cli):
    logger.debug(f"Executing: {' '.join(shell_cli)}")
    print(f"Executing: {' '.join(shell_cli)}")
    exec = subprocess.Popen(shell_cli, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    output, err = exec.communicate()
    print(output.decode("UTF-8"))
    print(err.decode("UTF-8"))


import argparse
import configparser
from logging import getLogger
import os
import uuid
import signal
import shutil
import sys
from functools import partial
from .base import (
    download_file,
)
from .iso import (
    create_grub_cfg,
    iso_volume_id,
    create_kickstarted_iso,
)
from .constants import (
    grub_cfg_template,
    isolinux_cfg_template
)

logger = getLogger(__name__)
logger.setLevel("DEBUG")


def cli_arguments():
    parser = argparse.ArgumentParser(prog="kiso", description="Add kickstart files to rhel based iso images")
    parser.add_argument(
            "--work-dir",
            default="/tmp",
            type=str,
            help="Working directory where to store temp files"
        )
    parser.add_argument(
            "-c",
            "--config",
            type=str,
            default="/etc/kiso/config.ini",
            help="Config file path"
        )
    parser.add_argument(
            "-b",
            "--build",
            help="Image to build"
        )
    parser.add_argument(
            "--remove-source",
            action='store_true',
            help="Remove source iso file"
        )
    return parser.parse_args()


def config_parser(conf_file):
    logger.debug(f"Parsing config file: {conf_file}")
    config = configparser.ConfigParser()
    with open(conf_file, "r") as file:
        config.read_string(file.read())
    return config


def handler(working_path, signum, frame):
    print()
    logger.error("CTR+C was pressed, cleaning up")
    sys.exit(1)


def cleanup(working_path, source_iso=None):
    logger.debug(f"Removing: {working_path}")
    shutil.rmtree(working_path)
    if source_iso:
        logger.debug(f"Removing: {source_iso}")
        os.remove(source_iso)


def process_parent(config, arguments):
    parent_ks_section = config.get(arguments.build, "parent")
    logger.debug(f"Parent: {parent_ks_section}")
    parent_url = config.get(parent_ks_section, "url")
    logger.debug(f"Parent download url: {parent_url}")
    parent_ks = config.get(parent_ks_section, "kcfg")
    logger.debug(f"Parent kcfg: {parent_ks}")
    child_ks = config.get(arguments.build, "kcfg")
    logger.debug(f"Child kcfg: {child_ks}")
    return parent_url, parent_ks, child_ks


def main():
    arguments = cli_arguments()
    try:
        work_subdir = uuid.uuid4().hex
        child_ks = None
        working_path = os.path.join(arguments.work_dir, work_subdir)
        signal.signal(signal.SIGINT, partial(handler, working_path))
        logger.debug(f"Argument selection: {arguments}")
        logger.debug(f"Working path: {working_path}")
        if os.path.exists(arguments.config):
            kiso_name = f"{arguments.build}-kickstarted.iso"
            config = config_parser(arguments.config)
            iso_save_path = os.path.join(config.get("DEFAULT", "storage"))
            updated_iso_path = os.path.join(iso_save_path, kiso_name)
            if not os.path.isdir(iso_save_path):
                logger.warning(f"Iso save path doesn't exists, creating {iso_save_path}")
                os.makedirs(iso_save_path)
        if arguments.build in config.sections():
            print(f"Building: {arguments.build}")
            if config.has_option(arguments.build, "parent"):
                iso_url, k_cfg, child_ks = process_parent(config, arguments)
                iso_file_name = f"{config.get(arguments.build, 'parent')}.iso"
                logger.debug("Child ks: %s", child_ks)
            else:
                iso_url = config.get(arguments.build, "url")
                k_cfg = config.get(arguments.build, "kcfg")
                iso_file_name = f"{arguments.build}.iso"
            logger.debug("Downloading iso from: %s", iso_url)
            iso_download_location = os.path.join(arguments.work_dir, iso_file_name)
        iso_location = download_file(iso_url, iso_download_location)
        os.makedirs(working_path)
        print(f"working path: {working_path}")
        iso_file = os.path.join(arguments.work_dir, iso_file_name)
        iso_volume_name = iso_volume_id(iso_file)
        logger.debug("Iso volume name: %s", iso_volume_name)
        efi_boot_cfg = create_grub_cfg(iso_volume_name, grub_cfg_template, arguments.build)
        grub_menu = os.path.join(working_path, "grub.cfg")
        print(f"Creating grub.cfg for {arguments.build} at {grub_menu}")
        with open(grub_menu, "w") as grub_file:
            print(f"writing grub.cfg to {os.path.join(working_path, '/grub.cfg')}")
            grub_file.write(efi_boot_cfg)
        iso_default_boot_file = "boot/grub2/grub.cfg"
        if arguments.build == "rocky-10":
            print("Detected rocky-10 build, using efi boot config")
            default_boot_menu = grub_menu
        else:
            print("Using isolinux boot config")
            grub_cfg = create_grub_cfg(iso_volume_name, isolinux_cfg_template, arguments.build)
            iso_default_boot_file = "isolinux/isolinux.cfg"
            default_boot_menu = os.path.join(working_path, "isolinux.cfg")
            with open(default_boot_menu, "w") as isolinux_file:
                print(f"writing isolinux.cfg to {os.path.join(working_path, '/isolinux.cfg')}")
                isolinux_file.write(grub_cfg)
        logger.info("Using kickstart file: %s", k_cfg)
        dst_iso=os.path.join(working_path, "kickstarted.iso")
        logger.debug("Destination iso: %s", dst_iso)
        if config.has_option(arguments.build, "parent"):
            modified_kcfg = os.path.join(working_path, "ks.cfg")
            shutil.copy(k_cfg, modified_kcfg)
            logger.debug("Appending child kickstart file: %s", child_ks)
            with open(modified_kcfg, "a") as append:
                append.write(f"\n%ksappend /run/install/repo/{os.path.basename(child_ks)}\n")
            logger.info("Creating kickstarted iso image")
            create_kickstarted_iso(
                src=iso_file,
                dst=dst_iso,
                iso_name=iso_volume_name,
                kickstart_cfg=modified_kcfg,
                grub_cfg=grub_menu,
                default_boot_cfg=default_boot_menu,
                default_boot_file=iso_default_boot_file,
                optional_kickstart_cfg=child_ks
            )
        else:
            create_kickstarted_iso(
                src=iso_file,
                dst=dst_iso,
                iso_name=iso_volume_name,
                kickstart_cfg=k_cfg,
                grub_cfg=grub_menu,
                default_boot_cfg=default_boot_menu,
                default_boot_file=iso_default_boot_file
            )
        logger.info("Moving %s iso to %s", dst_iso, updated_iso_path)
        os.rename(dst_iso, updated_iso_path)
    finally:
        logger.info("=" * 50)
        logger.info("Cleaninig up.")
        if os.path.isdir(working_path):
            if arguments.remove_source:
                logger.debug("Removing working dir and source iso")
                cleanup(working_path, iso_download_location)
            else:
                logger.debug("Removing working dir")
                cleanup(working_path)
